use flax::{entity_ids, BoxedSystem, Query, QueryBorrow, Schedule, System, World};
use glam::{
    f32::{Mat3, Vec3},
    Mat4,
};
use std::sync::Arc;
use vulkano::{
    buffer::{
        allocator::{SubbufferAllocator, SubbufferAllocatorCreateInfo},
        Buffer, BufferCreateInfo, BufferUsage,
    },
    command_buffer::{
        allocator::{CommandBufferAllocator, StandardCommandBufferAllocator},
        CommandBufferBeginInfo, CommandBufferLevel, CommandBufferUsage, RecordingCommandBuffer,
        RenderPassBeginInfo,
    },
    descriptor_set::{
        allocator::StandardDescriptorSetAllocator, DescriptorSet, WriteDescriptorSet
    },
    device::DeviceOwned,
    format::Format,
    image::{view::ImageView, Image, ImageCreateInfo, ImageType, ImageUsage},
    memory::allocator::{AllocationCreateInfo, MemoryTypeFilter, StandardMemoryAllocator},
    pipeline::{graphics::viewport::Viewport, GraphicsPipeline, Pipeline, PipelineBindPoint},
    render_pass::{Framebuffer, RenderPass},
    shader::EntryPoint,
    sync::GpuFuture,
};
use vulkano_util::{
    context::VulkanoContext, renderer::VulkanoWindowRenderer, window::VulkanoWindows,
};

use crate::{node_editor, render_module::RenderModule as ThreadLocalModule, vulkan::renderpass::make_render_pass};
use crate::{
    debug_gui::DebugGuiStack,
    vulkan::{
        framebuffer::make_framebuffer,
        pipeline::{make_pipeline, PassInfo, PipelineInfo},
    },
};
use egui_vulkano::Gui;
use super::node_editor::DemoApp;

use crate::temp::model::{INDICES, NORMALS, POSITIONS};

pub struct RenderModule {
    schedule: Schedule,
    memory_allocator: Arc<StandardMemoryAllocator>,
    descriptor_set_allocator: Arc<StandardDescriptorSetAllocator>,
    command_buffer_allocator: Arc<dyn CommandBufferAllocator>,
    viewport: Viewport,
    rotation_start: std::time::Instant,
    node_editor: DemoApp,
}

impl RenderModule {
    pub fn new(
        vk_context: &mut VulkanoContext,
        _vk_windows: &mut VulkanoWindows,
        _schedule: &mut Schedule,
        _world: &mut World,
        _events: &mut khors_core::events::Events,
    ) -> Self {
        let schedule = Schedule::builder()
            .with_system(add_distance_system())
            .build();

        let memory_allocator = Arc::new(StandardMemoryAllocator::new_default(
            vk_context.device().clone(),
        ));

        let descriptor_set_allocator = Arc::new(StandardDescriptorSetAllocator::new(
            vk_context.device().clone(),
            Default::default(),
        ));

        let command_buffer_allocator = Arc::new(StandardCommandBufferAllocator::new(
            vk_context.device().clone(),
            Default::default(),
        ));

        let viewport = Viewport {
            offset: [0.0, 0.0],
            extent: [0.0, 0.0],
            depth_range: 0.0..=1.0,
        };

        let node_editor = DemoApp::new();

        let rotation_start = std::time::Instant::now();

        Self {
            schedule,
            memory_allocator,
            descriptor_set_allocator,
            command_buffer_allocator,
            viewport,
            rotation_start,
            node_editor,
        }
    }
}

impl ThreadLocalModule for RenderModule {
    fn on_update(
        &mut self,
        gui_stack: &mut DebugGuiStack,
        vk_context: &mut VulkanoContext,
        vk_windows: &mut vulkano_util::window::VulkanoWindows,
        world: &mut World,
        _events: &mut khors_core::events::Events,
        _frame_time: std::time::Duration,
    ) -> anyhow::Result<()> {
        self.schedule.execute_seq(world).unwrap();

        let viewport = &mut self.viewport;

        for (window_id, renderer) in vk_windows.iter_mut() {
            let gui = gui_stack.get_mut(*window_id).unwrap();
            let node_editor = &mut self.node_editor;
            draw(
                vk_context.device().clone(),
                self.memory_allocator.clone(),
                self.descriptor_set_allocator.clone(),
                self.command_buffer_allocator.clone(),
                viewport,
                vk_context,
                renderer,
                gui,
                node_editor,
                self.rotation_start,
            );
        }
        Ok(())
    }
}

pub fn add_distance_system() -> BoxedSystem {
    let query = Query::new(entity_ids());

    System::builder()
        .with_query(query)
        .build(|mut query: QueryBorrow<'_, flax::EntityIds, _>| {
            for _id in &mut query {
                // println!("----------: {}", _id.index());
            }
        })
        .boxed()
}

#[allow(clippy::too_many_arguments)]
fn draw(
    device: Arc<vulkano::device::Device>,
    memory_allocator: Arc<StandardMemoryAllocator>,
    descriptor_set_allocator: Arc<StandardDescriptorSetAllocator>,
    command_buffer_allocator: Arc<dyn CommandBufferAllocator>,
    _viewport: &mut Viewport,
    context: &mut VulkanoContext,
    renderer: &mut VulkanoWindowRenderer,
    gui: &mut Gui,
    node_editor: &mut DemoApp,
    rotation_start: std::time::Instant,
) {
    let vertex_buffer = Buffer::from_iter(
        memory_allocator.clone(),
        BufferCreateInfo {
            usage: BufferUsage::VERTEX_BUFFER,
            ..Default::default()
        },
        AllocationCreateInfo {
            memory_type_filter: MemoryTypeFilter::PREFER_DEVICE
                | MemoryTypeFilter::HOST_SEQUENTIAL_WRITE,
            ..Default::default()
        },
        POSITIONS,
    )
    .unwrap();
    let normals_buffer = Buffer::from_iter(
        memory_allocator.clone(),
        BufferCreateInfo {
            usage: BufferUsage::VERTEX_BUFFER,
            ..Default::default()
        },
        AllocationCreateInfo {
            memory_type_filter: MemoryTypeFilter::PREFER_DEVICE
                | MemoryTypeFilter::HOST_SEQUENTIAL_WRITE,
            ..Default::default()
        },
        NORMALS,
    )
    .unwrap();
    let index_buffer = Buffer::from_iter(
        memory_allocator.clone(),
        BufferCreateInfo {
            usage: BufferUsage::INDEX_BUFFER,
            ..Default::default()
        },
        AllocationCreateInfo {
            memory_type_filter: MemoryTypeFilter::PREFER_DEVICE
                | MemoryTypeFilter::HOST_SEQUENTIAL_WRITE,
            ..Default::default()
        },
        INDICES,
    )
    .unwrap();

    let uniform_buffer = SubbufferAllocator::new(
        memory_allocator.clone(),
        SubbufferAllocatorCreateInfo {
            buffer_usage: BufferUsage::UNIFORM_BUFFER,
            memory_type_filter: MemoryTypeFilter::PREFER_DEVICE
                | MemoryTypeFilter::HOST_SEQUENTIAL_WRITE,
            ..Default::default()
        },
    );

    // let render_pass = vulkano::single_pass_renderpass!(
    //     device.clone(),
    //     attachments: {
    //         color: {
    //             format: renderer.swapchain_format(),
    //             samples: 1,
    //             load_op: Clear,
    //             store_op: Store,
    //         },
    //         depth_stencil: {
    //             format: Format::D16_UNORM,
    //             samples: 1,
    //             load_op: Clear,
    //             store_op: DontCare,
    //         },
    //     },
    //     pass: {
    //         color: [color],
    //         depth_stencil: {depth_stencil},
    //     },
    // )
    //     .unwrap();

    let render_pass = make_render_pass(device.clone(), renderer);

    let vs = vs::load(device.clone())
        .unwrap()
        .entry_point("main")
        .unwrap();
    let fs = fs::load(device.clone())
        .unwrap()
        .entry_point("main")
        .unwrap();

    let (mut pipeline, mut framebuffers) = window_size_dependent_setup(
        memory_allocator.clone(),
        vs.clone(),
        fs.clone(),
        renderer.swapchain_image_views(),
        render_pass.clone(),
    );

    // Do not draw the frame when the screen size is zero. On Windows, this can
    // occur when minimizing the application.
    let image_extent: [u32; 2] = renderer.window().inner_size().into();

    if image_extent.contains(&0) {
        return;
    }

    // Begin rendering by acquiring the gpu future from the window renderer.
    let previous_frame_end = renderer
        .acquire(None, |swapchain_images| {
            // Whenever the window resizes we need to recreate everything dependent
            // on the window size. In this example that
            // includes the swapchain, the framebuffers
            // and the dynamic state viewport.
            let (new_pipeline, new_framebuffers) = window_size_dependent_setup(
                memory_allocator.clone(),
                vs.clone(),
                fs.clone(),
                swapchain_images,
                render_pass.clone(),
            );

            pipeline = new_pipeline;
            framebuffers = new_framebuffers;
        })
        .unwrap();

    let uniform_buffer_subbuffer = {
        let elapsed = rotation_start.elapsed();
        let rotation = elapsed.as_secs() as f64 + elapsed.subsec_nanos() as f64 / 1_000_000_000.0;
        let rotation = Mat3::from_rotation_y(rotation as f32);

        // NOTE: This teapot was meant for OpenGL where the origin is at the lower left
        // instead the origin is at the upper left in Vulkan, so we reverse the Y axis.
        let aspect_ratio = renderer.aspect_ratio();

        let proj = Mat4::perspective_rh_gl(std::f32::consts::FRAC_PI_2, aspect_ratio, 0.01, 100.0);
        let view = Mat4::look_at_rh(
            Vec3::new(0.4, 0.3, 1.0),
            Vec3::new(0.0, 0.0, 0.0),
            Vec3::new(0.0, -1.0, 0.0),
        );
        let scale = Mat4::from_scale(Vec3::splat(0.01));

        let uniform_data = vs::Data {
            world: Mat4::from_mat3(rotation).to_cols_array_2d(),
            view: (view * scale).to_cols_array_2d(),
            proj: proj.to_cols_array_2d(),
        };

        let subbuffer = uniform_buffer.allocate_sized().unwrap();
        *subbuffer.write().unwrap() = uniform_data;

        subbuffer
    };

    let mut builder = RecordingCommandBuffer::new(
        command_buffer_allocator.clone(),
        context.graphics_queue().queue_family_index(),
        CommandBufferLevel::Primary,
        CommandBufferBeginInfo {
            usage: CommandBufferUsage::OneTimeSubmit,
            ..Default::default()
        },
    )
    .unwrap();

    let layout = &pipeline.layout().set_layouts()[0];
    let set = DescriptorSet::new(
        descriptor_set_allocator.clone(),
        layout.clone(),
        [WriteDescriptorSet::buffer(0, uniform_buffer_subbuffer)],
        [],
    )
    .unwrap();

    builder
        .begin_render_pass(
            RenderPassBeginInfo {
                clear_values: vec![Some([0.0, 0.0, 1.0, 1.0].into()), Some(1f32.into())],
                ..RenderPassBeginInfo::framebuffer(
                    framebuffers[renderer.image_index() as usize].clone(),
                )
            },
            Default::default(),
        )
        .unwrap()
        .bind_pipeline_graphics(pipeline.clone())
        .unwrap()
        .bind_descriptor_sets(
            PipelineBindPoint::Graphics,
            pipeline.layout().clone(),
            0,
            set,
        )
        .unwrap()
        .bind_vertex_buffers(0, (vertex_buffer.clone(), normals_buffer.clone()))
        .unwrap()
        .bind_index_buffer(index_buffer.clone())
        .unwrap();

    unsafe {
        builder
            .draw_indexed(index_buffer.len() as u32, 1, 0, 0, 0)
            .unwrap();
    }

    builder.end_render_pass(Default::default()).unwrap();
    // Finish recording the command buffer by calling `end`.
    let command_buffer = builder.end().unwrap();

    draw_gui(gui, node_editor);

    let before_future = previous_frame_end
        .then_execute(context.graphics_queue().clone(), command_buffer)
        .unwrap()
        .boxed();

    let after_future = gui
        .draw_on_image(before_future, renderer.swapchain_image_view())
        .boxed();

    // The color output is now expected to contain our triangle. But in order to
    // show it on the screen, we have to *present* the image by calling
    // `present` on the window renderer.
    //
    // This function does not actually present the image immediately. Instead it
    // submits a present command at the end of the queue. This means that it will
    // only be presented once the GPU has finished executing the command buffer
    // that draws the triangle.
    renderer.present(after_future, true);
}

fn draw_gui(gui: &mut Gui, node_editor: &mut DemoApp) {
    let mut code = CODE.to_owned();
    
    gui.immediate_ui(|gui| {
        let ctx = gui.context();

        node_editor.update(&ctx);

        egui::Window::new("Colors").vscroll(true).show(&ctx, |ui| {
            ui.vertical_centered(|ui| {
                ui.add(egui::widgets::Label::new("Hi there!"));
                sized_text(ui, "Rich Text", 32.0);
            });
            ui.separator();
            ui.columns(2, |columns| {
                egui::ScrollArea::vertical()
                    .id_source("source")
                    .show(&mut columns[0], |ui| {
                        ui.add(
                            egui::TextEdit::multiline(&mut code).font(egui::TextStyle::Monospace),
                        );
                    });
                egui::ScrollArea::vertical()
                    .id_source("rendered")
                    .show(&mut columns[1], |ui| {
                        ui.add(egui::widgets::Label::new("Good day!"));
                    });
            });
        });
    });
}

fn sized_text(ui: &mut egui::Ui, text: impl Into<String>, size: f32) {
    ui.label(
        egui::RichText::new(text)
            .size(size)
            .family(egui::FontFamily::Monospace),
    );
}

const CODE: &str = r"
# Some markup
```
let mut gui = Gui::new(&event_loop, renderer.surface(), None, renderer.queue(), SampleCount::Sample1);
```
";

fn window_size_dependent_setup(
    memory_allocator: Arc<StandardMemoryAllocator>,
    vs: EntryPoint,
    fs: EntryPoint,
    image_views: &[Arc<ImageView>],
    render_pass: Arc<RenderPass>,
) -> (Arc<GraphicsPipeline>, Vec<Arc<Framebuffer>>) {
    let device = memory_allocator.device().clone();

    let extent = image_views[0].image().extent();

    let depth_buffer = ImageView::new_default(
        Image::new(
            memory_allocator,
            ImageCreateInfo {
                image_type: ImageType::Dim2d,
                format: Format::D16_UNORM,
                extent,
                usage: ImageUsage::DEPTH_STENCIL_ATTACHMENT | ImageUsage::TRANSIENT_ATTACHMENT,
                ..Default::default()
            },
            AllocationCreateInfo::default(),
        )
        .unwrap(),
    )
    .unwrap();

    let framebuffers = image_views
        .iter()
        .map(|image_view| {
            make_framebuffer(
                render_pass.clone(),
                vec![image_view.clone(), depth_buffer.clone()],
            )
            .unwrap()
        })
        .collect::<Vec<_>>();

    // In the triangle example we use a dynamic viewport, as its a simple example. However in the
    // teapot example, we recreate the pipelines with a hardcoded viewport instead. This allows the
    // driver to optimize things, at the cost of slower window resizes.
    // https://computergraphics.stackexchange.com/questions/5742/vulkan-best-way-of-updating-pipeline-viewport
    let pipeline_info = PipelineInfo { vs, fs };

    let pass_info = PassInfo::new(render_pass.clone(), 0, extent);

    let pipeline = make_pipeline(device.clone(), &pipeline_info, &pass_info);

    (pipeline, framebuffers)
}

mod vs {
    vulkano_shaders::shader! {
        ty: "vertex",
        path: "src/temp/vert.glsl",
    }
}

mod fs {
    vulkano_shaders::shader! {
        ty: "fragment",
        path: "src/temp/frag.glsl",
    }
}

