use std::sync::Arc;

use vulkano::{device::Device, format::Format, render_pass::{AttachmentDescription, AttachmentReference, RenderPass as VkRenderPass, RenderPassCreateFlags, RenderPassCreateInfo, SubpassDependency, SubpassDescription, SubpassDescriptionFlags}};
use vulkano_util::renderer::VulkanoWindowRenderer;

#[derive(Default, Debug)]
pub struct SubpassInfo {
    pub color_attachments: Vec<Option<AttachmentReference>>,
    pub color_resolve_attachments: Vec<Option<AttachmentReference>>,
    pub input_attachments: Vec<Option<AttachmentReference>>,
    pub depth_stencil_attachment: Option<AttachmentReference>,
}

impl From<SubpassInfo> for SubpassDescription {
    fn from(value: SubpassInfo) -> Self {
        SubpassDescription {
            flags: SubpassDescriptionFlags::default(),
            color_attachments: value.color_attachments,
            color_resolve_attachments: value.color_resolve_attachments,
            input_attachments: value.input_attachments,
            depth_stencil_attachment: value.depth_stencil_attachment,
            ..Default::default()
        }
    }
}

fn make_subpass_description(subpass_info: Arc<SubpassInfo>) -> SubpassDescription {
    SubpassDescription {
        flags: SubpassDescriptionFlags::default(),
        color_attachments: subpass_info.color_attachments.clone(),
        color_resolve_attachments: subpass_info.color_resolve_attachments.clone(),
        input_attachments: subpass_info.input_attachments.clone(),
        depth_stencil_attachment: subpass_info.depth_stencil_attachment.clone(),
        ..Default::default()
    }
}

#[derive(Default)]
pub struct RenderPassInfo {
    pub attachments: Vec<AttachmentDescription>,
    pub subpasses: Vec<Arc<SubpassInfo>>,
    pub dependencies: Vec<SubpassDependency>,
}

pub struct RenderPass {
    device: Arc<Device>,
    render_pass: Arc<VkRenderPass>
}

impl RenderPass {
    pub fn new(device: Arc<vulkano::device::Device>, info: Arc<RenderPassInfo>) -> Self {
        let vk_subpasses = info.subpasses.iter().map(|subpass| make_subpass_description(subpass.clone())).collect::<Vec<SubpassDescription>>();
        let render_pass_create_info = RenderPassCreateInfo {
            attachments: info.attachments.clone(),
            dependencies: info.dependencies.clone(),
            subpasses: vk_subpasses,
            ..Default::default()
        };

        let render_pass = VkRenderPass::new(device.clone(), render_pass_create_info).unwrap();

        Self {
            device: device.clone(),
            render_pass
        }
    }

    pub fn renderpass(&self) -> Arc<VkRenderPass> {
        self.render_pass.clone()
    }
}

pub fn make_render_pass(device: Arc<vulkano::device::Device>, renderer: &mut VulkanoWindowRenderer) -> Arc<VkRenderPass> {
    let color_attachment = AttachmentDescription {
        format: renderer.swapchain_format(),
        samples: vulkano::image::SampleCount::Sample1,
        load_op: vulkano::render_pass::AttachmentLoadOp::Clear,
        store_op: vulkano::render_pass::AttachmentStoreOp::Store,
        initial_layout: vulkano::image::ImageLayout::ColorAttachmentOptimal,
        final_layout: vulkano::image::ImageLayout::ColorAttachmentOptimal,
        ..Default::default()
    };
    let depth_stencil_attachment = AttachmentDescription {
        format: Format::D16_UNORM,
        samples: vulkano::image::SampleCount::Sample1,
        load_op: vulkano::render_pass::AttachmentLoadOp::Clear,
        store_op: vulkano::render_pass::AttachmentStoreOp::DontCare,
        initial_layout: vulkano::image::ImageLayout::DepthStencilAttachmentOptimal,
        final_layout: vulkano::image::ImageLayout::DepthStencilAttachmentOptimal,
        ..Default::default()
    };

    let color_attachment_reference = AttachmentReference {
        attachment: 0,
        layout: color_attachment.initial_layout,
        ..Default::default()
    };

    let depth_stencil_attachment_reference = AttachmentReference {
        attachment: 1,
        layout: depth_stencil_attachment.initial_layout,
        ..Default::default()
    };
    
    let subpass_description = SubpassDescription {
        color_attachments: vec![Some(color_attachment_reference)],
        depth_stencil_attachment: Some(depth_stencil_attachment_reference),
        ..Default::default()
    };
    
    VkRenderPass::new(device.clone(), RenderPassCreateInfo {
        attachments: vec![color_attachment, depth_stencil_attachment],
        subpasses: vec![subpass_description],
        ..Default::default()
    }).unwrap()
}
